Package: BSgenome
Title: Software infrastructure for efficient representation of full
        genomes and their SNPs
Description: Infrastructure shared by all the Biostrings-based genome data
	packages.
biocViews: Genetics, Infrastructure, DataRepresentation,
        SequenceMatching, Annotation, SNP
URL: https://bioconductor.org/packages/BSgenome
BugReports: https://github.com/Bioconductor/BSgenome/issues
Version: 1.81.0
License: Artistic-2.0
Encoding: UTF-8
Authors@R: person("Hervé", "Pagès", role=c("aut", "cre"),
		  email="hpages.on.github@gmail.com")
Depends: R (>= 2.8.0), methods, BiocGenerics (>= 0.13.8), S4Vectors (>=
        0.47.6), IRanges (>= 2.13.16), Seqinfo, GenomicRanges (>=
        1.61.1), Biostrings (>= 2.77.2), BiocIO, rtracklayer (>=
        1.69.1)
Imports: utils, stats, matrixStats, XVector, Rsamtools (>= 2.25.1)
Suggests: BiocManager, GenomeInfoDb, BSgenome.Celegans.UCSC.ce2,
        BSgenome.Hsapiens.UCSC.hg38,
        BSgenome.Hsapiens.UCSC.hg38.masked,
        BSgenome.Mmusculus.UCSC.mm10, BSgenome.Rnorvegicus.UCSC.rn5,
        BSgenome.Scerevisiae.UCSC.sacCer1,
        BSgenome.Hsapiens.NCBI.GRCh38,
        TxDb.Hsapiens.UCSC.hg38.knownGene,
        TxDb.Mmusculus.UCSC.mm10.knownGene,
        SNPlocs.Hsapiens.dbSNP144.GRCh38,
        XtraSNPlocs.Hsapiens.dbSNP144.GRCh38, hgu95av2probe, RUnit,
        BSgenomeForge
LazyLoad: yes
Collate: utils.R OnDiskLongTable_old-class.R OnDiskLongTable-class.R
        OnDiskNamedSequences-class.R SNPlocs-class.R
        ODLT_SNPlocs-class.R OldFashionSNPlocs-class.R
        InjectSNPsHandler-class.R XtraSNPlocs-class.R BSgenome-class.R
        available.genomes.R injectSNPs.R getSeq-methods.R
        extractAt-methods.R bsapply.R BSgenomeViews-class.R
        BSgenome-utils.R export-methods.R AdvancedBSgenomeForge.R
git_url: https://git.bioconductor.org/packages/BSgenome
git_branch: devel
git_last_commit: bc88595
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 21:37:15 UTC; biocbuild
Author: Hervé Pagès [aut, cre]
Maintainer: Hervé Pagès <hpages.on.github@gmail.com>
