[<img src="https://www.bioconductor.org/images/logo/jpg/bioconductor_logo_rgb.jpg" width="200" align="right"/>](https://bioconductor.org/)

**BSgenome** is an R/Bioconductor package that provides the software infrastructure for efficient representation of full genomes and their SNPs.

See https://bioconductor.org/packages/BSgenome for more information including how to install the release version of the package (please refrain from installing directly from GitHub).

